#include <stdio.h>
#include <stdlib.h>

#include <mpi.h>

#define PRINTMTX 0

int main (int argc, char ** argv) {

  int n, i, j;
  int size, rank, rlen;
  double ** d, * temp, * row;
  char myname[MPI_MAX_PROCESSOR_NAME]; 

  MPI_Init (&argc, &argv);
  MPI_Comm_size (MPI_COMM_WORLD, &size);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);
  MPI_Get_processor_name (myname, &rlen);

  fprintf (stdout, "I am proces %3d of %3d my name is %s (%s)\n", 
    rank, size, myname, argv[0]);

  fflush(stdout);

  n = size;

  if (rank ==0) {
    if ( (d = (double **) malloc (sizeof (double *) * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
      MPI_Abort (MPI_COMM_WORLD, 1);
      return EXIT_FAILURE;
    }
    else {
      if ( (temp = (double *) malloc (sizeof (double) * n * n)) == NULL) {
        fprintf (stderr, "Memory allocation error\n");
        MPI_Abort (MPI_COMM_WORLD, 1);
        return EXIT_FAILURE;
      }
      else {
        for (i=0; i<n; i++) 
          d[i] = &temp[i*n]; 
      }
    }

    for (i=0; i<n; i++) 
      for (j=0; j<n; j++)
        d[i][j] = (double) i;
  }

  if ( (row = (double *) malloc (sizeof (double) * n)) == NULL) {
    fprintf (stderr, "Memory allocation error\n");
    MPI_Abort (MPI_COMM_WORLD, 1);
    return EXIT_FAILURE;
  }

  MPI_Scatter (temp, n, MPI_DOUBLE, row, n, MPI_DOUBLE, 0, MPI_COMM_WORLD); 

  MPI_Barrier (MPI_COMM_WORLD);
  fprintf (stdout, "rank %d \n", rank);
  fflush (stdout);
  for (i=0; i<n; i++) 
    fprintf (stdout, "%10.5f ", row[i]);
  fprintf (stdout, "\n");
  fflush (stdout);
  MPI_Barrier (MPI_COMM_WORLD);

  if (rank == 0) {
    free (d);
    free (temp);
  }
  else
    free (row);

  MPI_Finalize ();

  return EXIT_SUCCESS;
}
